
df_show(args...) = Base.show(args...)

df_show_type(io::IO, ::Type{T}) where {T} = dfprintln(io, crayon"red", "$T:", crayon"reset")
df_show_type(io::IO, x) = df_show_type(io, typeof(x))

function df_show(io::IO, block::InputData)
    s = """Block name: $(block.name)
    Block option: $(block.option)
    Block data:
    """
    dfprintln(io, s)
    dfprintln(io, string(block.data) * "\n\n")
    return
end

function df_show(io::IO, band::Band)
    df_show_type(io, band)
    string = """
    k_points of length $(length(band.k_points_cart)):
    cart:    $(band.k_points_cart[1]) -> $(band.k_points_cart[end])
    """
    !isempty(band.k_points_cryst) &&
        (string *= "cryst:   $(band.k_points_cryst[1]) -> $(band.k_points_cryst[end])\n")
    string *= """eigvals: $(band.eigvals[1]) -> $(band.eigvals[end])
    extra:   $(band.extra)
    """
    dfprintln(io, string)
    return
end

function df_show(io::IO, job::Job)
    server = Server(job.server)
    reset = crayon"reset"
    fieldns = [:name, :version]
    fs = string.(filter(x -> !isempty(x), getfield.((job,), fieldns)))
    fns = string.(fieldns)
    insert!(fns, 2, "dir")
    push!(fns, "server")
    if Servers.isalive(server) 
        push!(fs, "$(job.server), alive")
    else
        push!(fs, "$(job.server), not alive")
    end
        
    
    insert!(fs, 2, Jobs.main_job_dir(job))
    push!(fns, "versions")
    if Servers.isalive(server)
        versions = Client.versions(job)
        push!(fs, join(string.(versions), ", "))
        timestamp = Client.submission_time(job)
        if timestamp != 0
            push!(fns, "last submission")
            push!(fs, string(round(Dates.unix2datetime(timestamp), Second)))
        end
        push!(fns, "state")
        state = Client.state(job)
        push!(fs, string(state))
    else
        push!(fs, "unknown")
        push!(fns, "state")
        push!(fs, string(Jobs.Unknown))
    end
        
    lfns = maximum(length.(fns)) + maximum(length.(fs)) + 4
    line = "+"
    for i in 1:div(lfns, 2) + 1
        line *= "-"
    end
    dfprint(io, crayon"cyan", line[1:end-2])
    dfprint(io, crayon"cyan", "JOB")
    for i in 1:div(lfns, 2)
        line *= "-"
    end
    line *= "+"
    totlen = length(line)
    dfprintln(io, crayon"cyan", line[div(lfns, 2)+6:end], reset)
    lfns = maximum(length.(string.(fns)))
    for (fn, f) in zip(fns, fs)
        isname = fn == "name"
        l = length(fn)
        fn *= ":"
        for i in 1:lfns-l
            fn *= " "
        end
        dfprint(io, crayon"cyan", "|", reset)
        isname ? dfprint(io, " $fn ", crayon"magenta", "$f", reset) :
        dfprint(io, " $fn $f", reset)
        for i in 1:totlen-(length(fn)+4+length(f))
            dfprint(io, " ")
        end
        dfprintln(io, crayon"cyan", "|", reset)
    end
    is = job.calculations
    last = Servers.isalive(server) && ispath(server, Jobs.main_job_dir(job)) ? Client.last_running_calculation(job) : -1
    if !isempty(is)
        dfprintln(io, crayon"cyan", line, reset)
        dfprintln(io, reset, "(", crayon"green", "scheduled", reset, ", ", crayon"red",
                  "not scheduled", reset, ")")
        ln = maximum(length.(string.(map(x -> x.name, is))))
        for i in is
            n = i.name
            cr = i.run ? crayon"green" : crayon"red"
            if i == last
                if state == Jobs.Running
                    dfprint(io, cr, "\t$n <- running\n")
                elseif state == Jobs.Completed || state == Jobs.Failed || state == Jobs.Cancelled
                    dfprint(io, cr, "\t$n <- ran last\n")
                else
                    dfprint(io, cr, "\t$n\n")
                end
            else
                dfprint(io, cr, "\t$n\n")
            end
        end
    end
    dfprint(io, reset)
    return
end

function write_flags(io, fls, prestr = "")
    flags = String[]
    vals  = String[]
    for (f, v) in fls
        if v isa AbstractArray
            for i in findall(!iszero, v)
                push!(flags, string(f) * "($(join(Tuple(i), ",")))")
                push!(vals, string(v[i]))
            end
        else
            push!(flags, string(f))
            push!(vals, string(v))
        end
    end
    fl = maximum(length, flags)
    sp = sortperm(flags, by=length)
    for s in sp
        fs, v = flags[s], vals[s]
        l = length(fs)
        for i in 1:fl-l
            fs *= " "
        end
        dfprintln(io, crayon"cyan", prestr * "  $fs", crayon"yellow", " => ",
                  crayon"magenta", v, crayon"reset")
    end
    dfprintln(io,"")
end

function df_show(io::IO, c::Calculation)
    df_show_type(io, c)
    s = """name  = $(c.name)
    exec = $(c.exec.exec)
    run   = $(c.run)
    data  = $([e.name for e in c.data])
    flags:"""
    dfprintln(io, s)

    if !isempty(c.flags)
        write_flags_separately = false
        for (b, bflags) in c.flags
            if bflags isa Dict
                if !isempty(bflags)
                    dfprintln(io, crayon"green", "  &$b", crayon"reset")
                    write_flags(io, bflags, "  ")
                    
                end
            else
                write_flags_separately = true
            end 
        end
        if write_flags_separately
            write_flags(io, c.flags)
        end
    end
    dfprint(io, crayon"reset")
    return
end

function df_show(io::IO, flag_info::Calculations.QEFlagInfo{T}) where {T}
    df_show_type(io, flag_info)
    dfprintln(io, crayon"yellow", "name : $(flag_info.name)")
    dfprintln(io, crayon"cyan", "description:", crayon"reset")
    dfprint(io, "\t" * replace(flag_info.description, "\n" => "\n\t"))
    return
end

function df_show(io::IO, flag_info::Calculations.ElkFlagInfo{T}) where {T}
    df_show_type(io, flag_info)
    dfprintln(io, crayon"yellow", "name : $(flag_info.name)")
    dfprintln(io, crayon"green", "default:", crayon"reset")
    if flag_info.default != nothing
        dfprintln(io, "\t" * "$(flag_info.default)")
    else
        dfprintln(io, "\t" * "nothing")
    end
    dfprintln(io, crayon"cyan", "description:", crayon"reset")
    dfprintln(io, "\t" * replace(flag_info.description, "\n" => "\n\t"))
    return
end

function df_show(io::IO, info::Calculations.ElkControlBlockInfo)
    dfprintln(io, crayon"yellow", "name = $(info.name)")
    dfprintln(io, crayon"cyan", "flags:", crayon"reset")
    for flag in info.flags
        dfprintln(io, "\t$(flag.name)")
    end
    dfprintln(io, crayon"cyan", "description:", crayon"reset")
    dfprintln(io, "\t" * replace(info.description, "\n" => "\n\t"))
    dfprint(io, crayon"reset")
    return
end

function df_show(io::IO, el::Element)
    for f in fieldnames(typeof(el))
        dfprintln(io, crayon"red", "$f: ", crayon"reset", "$(getfield(el, f))")
    end
    return dfprint(io, crayon"reset")
end

function df_show(io::IO, str::Structure)
    dfprintln(io, crayon"cyan", "Structure", crayon"reset")
    dfprintln(io, crayon"red", "    cell parameters:")
    dfprint(io, crayon"reset",
            "\t a = $((str.cell[:,1]...,))\n\t b = $((str.cell[:,2]...,))\n\t c = $((str.cell[:,3]...,))\n")
    dfprintln(io, crayon"red", "    nat:", crayon"reset", " $(length(str.atoms))")
    dfprintln(io, crayon"red", "    ntyp:", crayon"reset", " $(length(unique(str.atoms)))")
    show(io, str.atoms)
    return dfprintln(io, crayon"reset")
end

function df_show(io::IO, at::Atom)
    dfprintln(io)
    dfprintln(io, crayon"cyan", "Atom")
    dfprintln(io, crayon"red", "    name: ", crayon"reset", "$(at.name)")
    for f in fieldnames(typeof(at))[3:end-1]
        f == :pseudo && continue
        fld = getfield(at, f)
        if f in (:position_cart, :position_cryst)
            dfprintln(io, crayon"red", "    $f: ", crayon"reset", "$((fld...,))")
        else
            dfprintln(io, crayon"red", "    $f: ", crayon"reset", "$fld")
        end
    end
    dfprintln(io, crayon"red", "    dftu:", crayon"reset")
    df_show(io, at.dftu)
    dfprintln(io, crayon"reset")
    return
end

function df_show(io::IO, proj::Projection)
    println(io, crayon"cyan", "Orbital: ", crayon"reset", "$(proj.orbital.name)")
    println(io, crayon"red", "start index: ", crayon"reset", "$(proj.start)")
    return println(io, crayon"red", "last index: ", crayon"reset", "$(proj.last)")
end
